"""Common exceptions."""

__all__ = [
    'ConfigReadError',
    'SessionTimeout',
    'UserAbort',
    'WrongPassword'
]


class ConfigReadError(Exception):
    """Indicates an error while reading the configuration."""


class SessionTimeout(Exception):
    """Indicates that the session timed out."""


class UserAbort(Exception):
    """Indicates that a required action has been aborted by the user."""


class WrongPassword(Exception):
    """Indicates a wrong password."""



from typing import Any, Sequence

from lib.archive import *

WORK_BASE_TYPES = set(
    [
        # "EPalWorkableType::Illegal",
        "EPalWorkableType::Progress",
        # "EPalWorkableType::CollectItem",
        # "EPalWorkableType::TransportItem",
        "EPalWorkableType::TransportItemInBaseCamp",
        "EPalWorkableType::ReviveCharacter",
        # "EPalWorkableType::CollectResource",
        "EPalWorkableType::LevelObject",
        "EPalWorkableType::Repair",
        "EPalWorkableType::Defense",
        "EPalWorkableType::BootUp",
        "EPalWorkableType::OnlyJoin",
        "EPalWorkableType::OnlyJoinAndWalkAround",
        "EPalWorkableType::RemoveMapObjectEffect",
        "EPalWorkableType::MonsterFarm",
    ]
)


def decode(
    reader: FArchiveReader, type_name: str, size: int, path: str
) -> dict[str, Any]:
    if type_name != "ArrayProperty":
        raise Exception(f"Expected ArrayProperty, got {type_name}")
    value = reader.property(type_name, size, path, nested_caller_path=path)
    for work_element in value["value"]["values"]:
        work_bytes = work_element["RawData"]["value"]["values"]
        work_type = work_element["WorkableType"]["value"]["value"]
        work_element["RawData"]["value"] = decode_bytes(work_bytes, work_type)
        for work_assign in work_element["WorkAssignMap"]["value"]:
            work_assign_bytes = work_assign["value"]["RawData"]["value"]["values"]
            work_assign["value"]["RawData"]["value"] = decode_work_assign_bytes(
                work_assign_bytes
            )
    return value


def decode_bytes(b_bytes: Sequence[int], work_type: str) -> dict[str, Any]:
    reader = FArchiveReader(bytes(b_bytes), debug=False)
    data = {}
    # Handle base serialization
    if work_type in WORK_BASE_TYPES:
        data["id"] = reader.guid()
        data["workable_bounds"] = {
            "location": {
                "x": reader.double(),
                "y": reader.double(),
                "z": reader.double(),
            },
            "rotation": {
                "x": reader.double(),
                "y": reader.double(),
                "z": reader.double(),
                "w": reader.double(),
            },
            "box_sphere_bounds": {
                "origin": {
                    "x": reader.double(),
                    "y": reader.double(),
                    "z": reader.double(),
                },
                "box_extent": {
                    "x": reader.double(),
                    "y": reader.double(),
                    "z": reader.double(),
                },
                "sphere_radius": reader.double(),
            },
        }
        data["base_camp_id_belong_to"] = reader.guid()
        data["owner_map_object_model_id"] = reader.guid()
        data["owner_map_object_concrete_model_id"] = reader.guid()
        data["current_state"] = reader.byte()
        data["assign_locations"] = reader.tarray(
            lambda r: {
                "location": {
                    "x": r.double(),
                    "y": r.double(),
                    "z": r.double(),
                },
                "facing_direction": {
                    "x": r.double(),
                    "y": r.double(),
                    "z": r.double(),
                },
            }
        )
        data["behaviour_type"] = reader.byte()
        data["assign_define_data_id"] = reader.fstring()
        data["override_work_type"] = reader.byte()
        data["assignable_fixed_type"] = reader.byte()
        data["assignable_otomo"] = reader.u32() > 0
        data["can_trigger_worker_event"] = reader.u32() > 0
        data["can_steal_assign"] = reader.u32() > 0
        if work_type == "EPalWorkableType::Defense":
            data["defense_combat_type"] = reader.byte()
        elif work_type == "EPalWorkableType::Progress":
            data["required_work_amount"] = reader.float()
            data["work_exp"] = reader.i32()
            data["current_work_amount"] = reader.float()
            data["auto_work_self_amount_by_sec"] = reader.float()
        elif work_type == "EPalWorkableType::ReviveCharacter":
            data["target_individual_id"] = {
                "player_uid": reader.guid(),
                "instance_id": reader.guid(),
            }
    # These two do not serialize base data
    elif work_type in ["EPalWorkableType::Assign", "EPalWorkableType::LevelObject"]:
        data["handle_id"] = reader.guid()
        data["location_index"] = reader.i32()
        data["assign_type"] = reader.byte()
        data["assigned_individual_id"] = {
            "player_uid": reader.guid(),
            "instance_id": reader.guid(),
        }
        data["state"] = reader.byte()
        data["fixed"] = reader.u32()
        if work_type == "EPalWorkableType::LevelObject":
            data["target_map_object_model_id"] = reader.guid()

    if len(data.keys()) == 0:
        print(f"Warning, unable to parse {work_type}, falling back to raw bytes")
        return {"values": b_bytes}
    # UPalWorkProgressTransformBase->SerializeProperties
    transform_type = reader.byte()
    data["transform"] = {"type": transform_type}
    if transform_type == 1:
        data["transform"]["location"] = {
            "x": reader.double(),
            "y": reader.double(),
            "z": reader.double(),
        }
        data["transform"]["rotation"] = {
            "x": reader.double(),
            "y": reader.double(),
            "z": reader.double(),
            "w": reader.double(),
        }
        data["transform"]["scale"] = {
            "x": reader.double(),
            "y": reader.double(),
            "z": reader.double(),
        }
    elif transform_type == 2:
        data["transform"]["map_object_instance_id"] = reader.guid()
    elif transform_type == 3:
        data["transform"]["guid"] = reader.guid()
        data["transform"]["instance_id"] = reader.guid()
    else:
        remaining_data = reader.read_to_end()
        print(
            f"Unknown EPalWorkTransformType, please report this: {transform_type}: {work_type}: {''.join(f'{b:02x}' for b in remaining_data)}"
        )
        data["transform"]["raw_data"] = [b for b in remaining_data]

    if not reader.eof():
        raise Exception(
            f"Warning: EOF not reached for {work_type}, remaining bytes: {reader.read_to_end()}"
        )

    return data


def decode_work_assign_bytes(b_bytes: Sequence[int]) -> dict[str, Any]:
    reader = FArchiveReader(bytes(b_bytes), debug=False)
    data = {}

    data["id"] = reader.guid()
    data["location_index"] = reader.i32()
    data["assign_type"] = reader.byte()
    data["assigned_individual_id"] = {
        "player_uid": reader.guid(),
        "instance_id": reader.guid(),
    }
    data["state"] = reader.byte()
    data["fixed"] = reader.u32() > 0

    if not reader.eof():
        raise Exception("Warning: EOF not reached")

    return data


def encode(
    writer: FArchiveWriter, property_type: str, properties: dict[str, Any]
) -> int:
    if property_type != "ArrayProperty":
        raise Exception(f"Expected ArrayProperty, got {property_type}")
    del properties["custom_type"]
    for work_element in properties["value"]["values"]:
        work_type = work_element["WorkableType"]["value"]["value"]
        work_element["RawData"]["value"] = {
            "values": [
                b for b in encode_bytes(work_element["RawData"]["value"], work_type)
            ]
        }
        for work_assign in work_element["WorkAssignMap"]["value"]:
            work_assign["value"]["RawData"]["value"] = {
                "values": [
                    b
                    for b in encode_work_assign_bytes(
                        work_assign["value"]["RawData"]["value"]
                    )
                ]
            }
    return writer.property_inner(property_type, properties)


def encode_bytes(p: dict[str, Any], work_type: str) -> bytes:
    writer = FArchiveWriter()

    # Handle base serialization
    if work_type in WORK_BASE_TYPES:
        writer.guid(p["id"])
        writer.double(p["workable_bounds"]["location"]["x"])
        writer.double(p["workable_bounds"]["location"]["y"])
        writer.double(p["workable_bounds"]["location"]["z"])
        writer.double(p["workable_bounds"]["rotation"]["x"])
        writer.double(p["workable_bounds"]["rotation"]["y"])
        writer.double(p["workable_bounds"]["rotation"]["z"])
        writer.double(p["workable_bounds"]["rotation"]["w"])
        writer.double(p["workable_bounds"]["box_sphere_bounds"]["origin"]["x"])
        writer.double(p["workable_bounds"]["box_sphere_bounds"]["origin"]["y"])
        writer.double(p["workable_bounds"]["box_sphere_bounds"]["origin"]["z"])
        writer.double(p["workable_bounds"]["box_sphere_bounds"]["box_extent"]["x"])
        writer.double(p["workable_bounds"]["box_sphere_bounds"]["box_extent"]["y"])
        writer.double(p["workable_bounds"]["box_sphere_bounds"]["box_extent"]["z"])
        writer.double(p["workable_bounds"]["box_sphere_bounds"]["sphere_radius"])
        writer.guid(p["base_camp_id_belong_to"])
        writer.guid(p["owner_map_object_model_id"])
        writer.guid(p["owner_map_object_concrete_model_id"])
        writer.byte(p["current_state"])
        writer.tarray(
            lambda w, l: (
                w.double(l["location"]["x"]),
                w.double(l["location"]["y"]),
                w.double(l["location"]["z"]),
                w.double(l["facing_direction"]["x"]),
                w.double(l["facing_direction"]["y"]),
                w.double(l["facing_direction"]["z"]),
            ),
            p["assign_locations"],
        )
        writer.byte(p["behaviour_type"])
        writer.fstring(p["assign_define_data_id"])
        writer.byte(p["override_work_type"])
        writer.byte(p["assignable_fixed_type"])
        writer.u32(1 if p["assignable_otomo"] else 0)
        writer.u32(1 if p["can_trigger_worker_event"] else 0)
        writer.u32(1 if p["can_steal_assign"] else 0)
        if work_type == "EPalWorkableType::Defense":
            writer.byte(p["defense_combat_type"])
        elif work_type == "EPalWorkableType::Progress":
            writer.float(p["required_work_amount"])
            writer.i32(p["work_exp"])
            writer.float(p["current_work_amount"])
            writer.float(p["auto_work_self_amount_by_sec"])
        elif work_type == "EPalWorkableType::ReviveCharacter":
            writer.guid(p["target_individual_id"]["player_uid"])
            writer.guid(p["target_individual_id"]["instance_id"])
    # These two do not serialize base data
    elif work_type in ["EPalWorkableType::Assign", "EPalWorkableType::LevelObject"]:
        writer.guid(p["handle_id"])
        writer.i32(p["location_index"])
        writer.byte(p["assign_type"])
        writer.guid(p["assigned_individual_id"]["player_uid"])
        writer.guid(p["assigned_individual_id"]["instance_id"])
        writer.byte(p["state"])
        writer.u32(p["fixed"])
        if work_type == "EPalWorkableType::LevelObject":
            writer.guid(p["target_map_object_model_id"])

    # UPalWorkProgressTransformBase->SerializeProperties
    transform_type = p["transform"]["type"]
    writer.byte(transform_type)
    if transform_type == 1:
        writer.double(p["transform"]["location"]["x"])
        writer.double(p["transform"]["location"]["y"])
        writer.double(p["transform"]["location"]["z"])
        writer.double(p["transform"]["rotation"]["x"])
        writer.double(p["transform"]["rotation"]["y"])
        writer.double(p["transform"]["rotation"]["z"])
        writer.double(p["transform"]["rotation"]["w"])
        writer.double(p["transform"]["scale"]["x"])
        writer.double(p["transform"]["scale"]["y"])
        writer.double(p["transform"]["scale"]["z"])
    elif transform_type == 2:
        writer.guid(p["transform"]["map_object_instance_id"])
    elif transform_type == 3:
        writer.guid(p["transform"]["guid"])
        writer.guid(p["transform"]["instance_id"])
    else:
        print(
            f"Unknown EPalWorkTransformType, please report this: {transform_type}: {work_type}"
        )
        writer.bytes(p["transform"]["raw_data"])

    encoded_bytes = writer.bytes()
    return encoded_bytes


def encode_work_assign_bytes(p: dict[str, Any]) -> bytes:
    writer = FArchiveWriter()

    writer.guid(p["id"])
    writer.i32(p["location_index"])
    writer.byte(p["assign_type"])
    writer.guid(p["assigned_individual_id"]["player_uid"])
    writer.guid(p["assigned_individual_id"]["instance_id"])
    writer.byte(p["state"])
    writer.u32(1 if p["fixed"] else 0)

    encoded_bytes = writer.bytes()
    return encoded_bytes



"""An interactive RCON shell."""

from argparse import ArgumentParser, Namespace
from logging import INFO, basicConfig, getLogger
from pathlib import Path

from rcon import battleye, source
from rcon.readline import CommandHistory
from rcon.config import CONFIG_FILES, LOG_FORMAT, from_args
from rcon.console import PROMPT, rconcmd
from rcon.errorhandler import ErrorHandler


__all__ = ['get_args', 'main']


LOGGER = getLogger('rconshell')


def get_args() -> Namespace:
    """Parses and returns the CLI arguments."""

    parser = ArgumentParser(description='An interactive RCON shell.')
    parser.add_argument('server', nargs='?', help='the server to connect to')
    parser.add_argument(
        '-B', '--battleye', action='store_true',
        help='use BattlEye RCon instead of Source RCON'
    )
    parser.add_argument(
        '-c', '--config', type=Path, metavar='file', default=CONFIG_FILES,
        help='the configuration file'
    )
    parser.add_argument(
        '-p', '--prompt', default=PROMPT, metavar='PS1',
        help='the shell prompt'
    )
    return parser.parse_args()


def run() -> None:
    """Runs the RCON shell."""

    args = get_args()
    basicConfig(level=INFO, format=LOG_FORMAT)
    client_cls = battleye.Client if args.battleye else source.Client

    if args.server:
        host, port, passwd = from_args(args)
    else:
        host = port = passwd = None

    with CommandHistory(LOGGER):
        rconcmd(client_cls, host, port, passwd, prompt=args.prompt)


def main() -> int:
    """Runs the main script with exceptions handled."""

    with ErrorHandler(LOGGER) as handler:
        run()

    return handler.exit_code



#!/usr/bin/env python3
# Author: MagicBear
# License: MIT License

import glob, sys, os, datetime, zlib, subprocess
from operator import itemgetter, attrgetter

module_dir = os.path.dirname(os.path.abspath(__file__))
sys.path.insert(0, module_dir)
sys.path.insert(0, os.path.join(module_dir, "../save_tools"))

from rcon.source import Client
import argparse
import json
from palworld_save_tools.gvas import GvasFile
from palworld_save_tools.palsav import compress_gvas_to_sav, decompress_sav_to_gvas
from palworld_save_tools.paltypes import PALWORLD_CUSTOM_PROPERTIES, PALWORLD_TYPE_HINTS
from palworld_save_tools.archive import *
from pathlib import Path

def main():
    global output_file, output_path, args, gui, playerMapping, instanceMapping

    parser = argparse.ArgumentParser(
        prog="palworld-playey-list",
        description="List player on the Players Folder",
    )
    parser.add_argument("filename", help="Filename of the player sav", nargs='?')
    parser.add_argument(
        "--host",
        "-H",
        default="127.0.0.1",
        help="Host for PalWorld Server RCON",
    )
    parser.add_argument(
        "--port",
        "-P",
        type=int,
        default=25575,
        help="PalWorld Server RCON Port",
    )
    parser.add_argument(
        "--password",
        "-p",
        help="RCON Password",
    )
    
    args = parser.parse_args()
    
    if not os.path.exists("%s/.config" % Path.home()):
        os.mkdir("%s/.config" % Path.home())
        
    if not os.path.exists("%s/.config/palworld-server-toolkit" % Path.home()):
        os.mkdir("%s/.config/palworld-server-toolkit" % Path.home())
    
    player_cache = "%s/.config/palworld-server-toolkit/player.txt" % Path.home()
    mapping = {}
    
    if os.path.exists(player_cache):
        with open(player_cache,"rb") as f:
            for line in f:
                try:
                    rs = line.decode("utf-8").strip().split(",")
                except UnicodeDecodeError as e:
                    continue
                if len(rs) == 3:
                    try:
                        mapping[int(rs[1])] = rs
                    except ValueError as e:
                        pass

    if args.password:
        client = Client(args.host, args.port, passwd=args.password, timeout=0.5)
        client.connect(True)
    
        try:
            players = client.run('ShowPlayers').split("\n")
            for player in players:
                player = player.split(",")
                if len(player) != 3:
                    break
                if player[0] == "name":
                    continue
                mapping[int(player[1])] = player
        except UnicodeDecodeError as e:
            pass

    with open(player_cache, "wb") as f:
        for k in mapping:
            f.write((",".join(mapping[k]) + "\n").encode('utf-8'))

    savlist = glob.glob("*.sav")
    lists = []
    for sav in savlist:
        lists.append({'mt': os.stat(sav).st_mtime, 'file': sav})

    lists.sort(key=itemgetter('mt'))
    for item in lists:
        sav = item['file']
        save_path = item['file']
        #  or int(sav[0:8], 16) not in mapping
        if args.filename is not None and args.filename in sav:
            module_dir = os.path.dirname(os.path.abspath(__file__))
            sys.path.insert(0, os.path.join(module_dir, "save_tools"))
            sys.path.insert(0, os.path.join(module_dir, "palworld-save-tools"))

            with open(item['file'], "rb") as f:
                # Read the file
                data = f.read()

                raw_gvas, _ = decompress_sav_to_gvas(data)
                player_gvas_file = GvasFile.read(raw_gvas, PALWORLD_TYPE_HINTS, PALWORLD_CUSTOM_PROPERTIES)
                PrettyPrint(player_gvas_file.properties['SaveData'])
                break
        elif args.filename is None:
            print("%s -> %d\t %s\t%s\t %s" % (sav[0:8], int(sav[0:8], 16),
                                              str(datetime.datetime.fromtimestamp(int(item['mt']))),
                                              mapping[int(sav[0:8], 16)][0] if int(sav[0:8], 16) in mapping else "",
                                              "https://steamcommunity.com/profiles/" + mapping[int(sav[0:8], 16)][
                                                  2] if int(sav[0:8], 16) in mapping else ""))


def PrettyPrint(data, level = 0):
    simpleType = ['DateTime', 'Guid', 'LinearColor', 'Quat', 'Vector', 'PalContainerId']
    if 'struct_type' in data:
        if data['struct_type'] == 'DateTime':
            print("%s<Value Type='DateTime'>%d</Value>" % ("  " * level, data['value']))
        elif data['struct_type'] == 'Guid':
            print("\033[96m%s\033[0m" % (data['value']), end="")
        elif data['struct_type'] == "LinearColor":
            print("%.f %.f %.f %.f" % (data['value']['r'],
                                                                           data['value']['g'],
                                                                           data['value']['b'],
                                                                           data['value']['a']), end="")
        elif data['struct_type'] == "Quat":
            print("%.f %.f %.f %.f" % (data['value']['x'],
                                                                           data['value']['y'],
                                                                           data['value']['z'],
                                                                           data['value']['w']), end="")
        elif data['struct_type'] == "Vector":
            print("%.f %.f %.f" % (data['value']['x'],
                                                                           data['value']['y'],
                                                                           data['value']['z']), end="")
        elif data['struct_type'] == "PalContainerId":
            print("\033[96m%s\033[0m" % (data['value']['ID']['value']), end="")
        elif isinstance(data['struct_type'], dict):
            print("%s<%s>" % ("  " * level, data['struct_type']))
            for key in data['value']:
                PrettyPrint(data['value'], level + 1)
            print("%s</%s>" % ("  " * level, data['struct_type']))
        else:
            PrettyPrint(data['value'], level + 1)
    else:
        for key in data:
            if not isinstance(data[key], dict):
                print("%s<%s type='unknow'>%s</%s>" % ("  " * level, key, data[key], key))
                continue
            if 'struct_type' in data[key] and data[key]['struct_type'] in simpleType:
                print("%s<%s type='%s'>" % ("  " * level, key, data[key]['struct_type']), end="")
                PrettyPrint(data[key], level + 1)
                print("</%s>" % (key))
            elif 'type' in data[key] and data[key]['type'] in ["IntProperty", "Int64Property", "BoolProperty"]:
                print("%s<%s Type='%s'>\033[95m%d\033[0m</%s>" % ("  " * level, key, data[key]['type'], data[key]['value'], key))
            elif 'type' in data[key] and data[key]['type'] == "FloatProperty":
                print("%s<%s Type='%s'>\033[95m%f\033[0m</%s>" % ("  " * level, key, data[key]['type'], data[key]['value'], key))
            elif 'type' in data[key] and data[key]['type'] in ["StrProperty", "ArrayProperty", "NameProperty"]:
                print("%s<%s Type='%s'>\033[95m%s\033[0m</%s>" % ("  " * level, key, data[key]['type'], data[key]['value'], key))
            elif isinstance(data[key], list):
                print("%s<%s Type='%s'>%s</%s>" % ("  " * level, key, data[key]['struct_type'] if 'struct_type' in data[
                    key] else "\033[31munknow struct\033[0m", str(data[key]), key))
            else:
                print("%s<%s Type='%s'>" % ("  " * level, key, data[key]['struct_type'] if 'struct_type' in data[key] else "\033[31munknow struct\033[0m"))
                PrettyPrint(data[key], level + 1)
                print("%s</%s>" % ("  " * level, key))


if __name__ == "__main__":
    main()



from typing import Any, Sequence

from lib.archive import *


def decode(
    reader: FArchiveReader, type_name: str, size: int, path: str
) -> dict[str, Any]:
    if type_name != "ArrayProperty":
        raise Exception(f"Expected ArrayProperty, got {type_name}")
    value = reader.property(type_name, size, path, nested_caller_path=path)
    data_bytes = value["value"]["values"]
    value["value"] = decode_bytes(data_bytes)
    return value


def connect_info_item_reader(reader: FArchiveReader) -> dict[str, Any]:
    return {
        "connect_to_model_instance_id": reader.guid(),
        "index": reader.byte(),
    }


def connect_info_item_writer(writer: FArchiveWriter, properties: dict[str, Any]):
    writer.guid(properties["connect_to_model_instance_id"])
    writer.byte(properties["index"])


def decode_bytes(c_bytes: Sequence[int]) -> dict[str, Any]:
    if len(c_bytes) == 0:
        return None
    reader = FArchiveReader(bytes(c_bytes), debug=False)
    data = {}
    data["supported_level"] = reader.i32()
    data["connect"] = {
        "index": reader.byte(),
        "any_place": reader.tarray(connect_info_item_reader),
    }
    # We are guessing here, we don't have information about the type without mapping object names -> types
    # Stairs have 2 connectors (up and down),
    # Roofs have 4 connectors (front, back, right, left)
    if not reader.eof():
        data["other_connectors"] = []
        while not reader.eof():
            data["other_connectors"].append(
                {
                    "index": reader.byte(),
                    "connect": reader.tarray(connect_info_item_reader),
                }
            )
        if len(data["other_connectors"]) not in [2, 4]:
            print(
                f"Warning: unknown connector type with {len(data['other_connectors'])} connectors"
            )
    return data


def encode(
    writer: FArchiveWriter, property_type: str, properties: dict[str, Any]
) -> int:
    if property_type != "ArrayProperty":
        raise Exception(f"Expected ArrayProperty, got {property_type}")
    del properties["custom_type"]
    encoded_bytes = encode_bytes(properties["value"])
    properties["value"] = {"values": [b for b in encoded_bytes]}
    return writer.property_inner(property_type, properties)


def encode_bytes(p: dict[str, Any]) -> bytes:
    if p is None:
        return bytes()
    writer = FArchiveWriter()
    writer.i32(p["supported_level"])
    writer.byte(p["connect"]["index"])
    writer.tarray(connect_info_item_writer, p["connect"]["any_place"])
    if "other_connectors" in p:
        for other in p["other_connectors"]:
            writer.byte(other["index"])
            writer.tarray(connect_info_item_writer, other["connect"])
    encoded_bytes = writer.bytes()
    return encoded_bytes



"""RCON exceptions."""

__all__ = ['RequestIdMismatch']


class RequestIdMismatch(Exception):
    """Indicates a mismatch in request IDs."""

    def __init__(self, sent: int, received: int):
        """Sets the IDs that have been sent and received."""
        super().__init__()
        self.sent = sent
        self.received = received



import io
import os
import struct
import uuid
from typing import Any, Callable, Optional, Union


class UUID:
    """Wrapper around uuid.UUID to delay evaluation of UUIDs until necessary"""

    __slots__ = ("raw_bytes", "parsed_uuid")
    raw_bytes: bytes
    parsed_uuid: uuid.UUID

    def __init__(self, raw_bytes: bytes) -> None:
        self.raw_bytes = raw_bytes
        self.parsed_uuid = None

    def from_str(s: str) -> "UUID":
        b = uuid.UUID(s).bytes
        return UUID(
            bytes(
                [
                    b[0x3],
                    b[0x2],
                    b[0x1],
                    b[0x0],
                    b[0x7],
                    b[0x6],
                    b[0x5],
                    b[0x4],
                    b[0xB],
                    b[0xA],
                    b[0x9],
                    b[0x8],
                    b[0xF],
                    b[0xE],
                    b[0xD],
                    b[0xC],
                ]
            )
        )

    def __str__(self) -> str:
        if not self.parsed_uuid:
            b = self.raw_bytes
            uuid_int = (
                b[0xC]
                + (b[0xD] << 8)
                + (b[0xE] << 16)
                + (b[0xF] << 24)
                + (b[0x8] << 32)
                + (b[0x9] << 40)
                + (b[0xA] << 48)
                + (b[0xB] << 56)
                + (b[0x4] << 64)
                + (b[0x5] << 72)
                + (b[0x6] << 80)
                + (b[0x7] << 88)
                + (b[0x0] << 96)
                + (b[0x1] << 104)
                + (b[0x2] << 112)
                + (b[0x3] << 120)
            )
            self.parsed_uuid = uuid.UUID(int=uuid_int)
        return str(self.parsed_uuid)

    def __eq__(self, __value: object) -> bool:
        return str(self) == str(__value)


def instance_id_reader(reader: "FArchiveReader") -> dict[str, UUID]:
    return {
        "guid": reader.guid(),
        "instance_id": reader.guid(),
    }


def uuid_reader(reader: "FArchiveReader") -> UUID:
    b = reader.read(16)
    if len(b) != 16:
        raise Exception("could not read 16 bytes for uuid")
    return UUID(b)


class FArchiveReader:
    data: io.BytesIO
    size: int
    type_hints: dict[str, str]
    custom_properties: dict[str, tuple[Callable, Callable]]
    debug: bool

    def __init__(
        self,
        data,
        type_hints: dict[str, str] = {},
        custom_properties: dict[str, tuple[Callable, Callable]] = {},
        debug: bool = os.environ.get("DEBUG", "0") == "1",
    ):
        self.data = io.BytesIO(data)
        self.size = len(data)
        self.type_hints = type_hints
        self.custom_properties = custom_properties
        self.debug = debug

    def __enter__(self):
        self.data.seek(0)
        return self

    def __exit__(self, type, value, traceback):
        self.data.close()

    def get_type_or(self, path: str, default: str):
        if path in self.type_hints:
            return self.type_hints[path]
        else:
            print(f"Struct type for {path} not found, assuming {default}")
            return default

    def eof(self) -> bool:
        return self.data.tell() >= self.size

    def read(self, size: int) -> bytes:
        return self.data.read(size)

    def read_to_end(self) -> bytes:
        return self.data.read(self.size - self.data.tell())

    def bool(self) -> bool:
        return self.byte() > 0

    def fstring(self) -> str:
        # in the hot loop, avoid function calls
        reader = self.data
        (size,) = self.unpack_i32(reader.read(4))

        if size == 0:
            return ""

        data: bytes
        encoding: str
        if size < 0:
            size = -size
            data = reader.read(size * 2)[:-2]
            encoding = "utf-16-le"
        else:
            data = reader.read(size)[:-1]
            encoding = "ascii"

        try:
            return data.decode(encoding)
        except Exception as e:
            try:
                escaped = data.decode(encoding, errors="surrogatepass")
                print(
                    f"Error decoding {encoding} string of length {size}, data loss may occur! {bytes(data)}"
                )
                return escaped
            except Exception as e:
                raise Exception(
                    f"Error decoding {encoding} string of length {size}: {bytes(data)}"
                ) from e

    unpack_i16 = struct.Struct("h").unpack

    def i16(self) -> int:
        return self.unpack_i16(self.data.read(2))[0]

    unpack_u16 = struct.Struct("H").unpack

    def u16(self) -> int:
        return self.unpack_u16(self.data.read(2))[0]

    unpack_i32 = struct.Struct("i").unpack

    def i32(self) -> int:
        return self.unpack_i32(self.data.read(4))[0]

    unpack_u32 = struct.Struct("I").unpack

    def u32(self) -> int:
        return self.unpack_u32(self.data.read(4))[0]

    unpack_i64 = struct.Struct("q").unpack

    def i64(self) -> int:
        return self.unpack_i64(self.data.read(8))[0]

    unpack_u64 = struct.Struct("Q").unpack

    def u64(self) -> int:
        return self.unpack_u64(self.data.read(8))[0]

    unpack_float = struct.Struct("f").unpack

    def float(self) -> float:
        return self.unpack_float(self.data.read(4))[0]

    unpack_double = struct.Struct("d").unpack

    def double(self) -> float:
        return self.unpack_double(self.data.read(8))[0]

    unpack_byte = struct.Struct("B").unpack

    def byte(self) -> int:
        return self.unpack_byte(self.data.read(1))[0]

    def byte_list(self, size: int) -> list[int]:
        return struct.unpack(str(size) + "B", self.data.read(size))

    def skip(self, size: int) -> None:
        self.data.read(size)

    def guid(self) -> UUID:
        # in the hot loop, avoid function calls
        return UUID(self.data.read(16))

    def optional_guid(self) -> Optional[UUID]:
        # in the hot loop, avoid function calls
        if self.data.read(1)[0]:
            return UUID(self.data.read(16))
        return None

    def tarray(
        self, type_reader: Callable[["FArchiveReader"], dict[str, Any]]
    ) -> list[dict[str, Any]]:
        count = self.u32()
        array = []
        for _ in range(count):
            array.append(type_reader(self))
        return array

    def properties_until_end(self, path: str = "") -> dict[str, Any]:
        properties = {}
        while True:
            name = self.fstring()
            if name == "None":
                break
            type_name = self.fstring()
            size = self.u64()
            properties[name] = self.property(type_name, size, f"{path}.{name}")
        return properties

    def property(
        self, type_name: str, size: int, path: str, nested_caller_path: str = ""
    ) -> dict[str, Any]:
        value = {}
        if path in self.custom_properties and (
            path is not nested_caller_path or nested_caller_path == ""
        ):
            value = self.custom_properties[path][0](self, type_name, size, path)
            value["custom_type"] = path
        elif type_name == "StructProperty":
            value = self.struct(path)
        elif type_name == "IntProperty":
            value = {
                "id": self.optional_guid(),
                "value": self.i32(),
            }
        elif type_name == "Int64Property":
            value = {
                "id": self.optional_guid(),
                "value": self.i64(),
            }
        elif type_name == "FixedPoint64Property":
            value = {
                "id": self.optional_guid(),
                "value": self.i32(),
            }
        elif type_name == "FloatProperty":
            value = {
                "id": self.optional_guid(),
                "value": self.float(),
            }
        elif type_name == "StrProperty":
            value = {
                "id": self.optional_guid(),
                "value": self.fstring(),
            }
        elif type_name == "NameProperty":
            value = {
                "id": self.optional_guid(),
                "value": self.fstring(),
            }
        elif type_name == "EnumProperty":
            enum_type = self.fstring()
            _id = self.optional_guid()
            enum_value = self.fstring()
            value = {
                "id": _id,
                "value": {
                    "type": enum_type,
                    "value": enum_value,
                },
            }
        elif type_name == "BoolProperty":
            value = {
                "value": self.bool(),
                "id": self.optional_guid(),
            }
        elif type_name == "ArrayProperty":
            array_type = self.fstring()
            value = {
                "array_type": array_type,
                "id": self.optional_guid(),
                "value": self.array_property(array_type, size - 4, path),
            }
        elif type_name == "MapProperty":
            key_type = self.fstring()
            value_type = self.fstring()
            _id = self.optional_guid()
            self.u32()
            count = self.u32()
            values = {}
            key_path = path + ".Key"
            if key_type == "StructProperty":
                key_struct_type = self.get_type_or(key_path, "Guid")
            else:
                key_struct_type = None
            value_path = path + ".Value"
            if value_type == "StructProperty":
                value_struct_type = self.get_type_or(value_path, "StructProperty")
            else:
                value_struct_type = None
            values = []
            for _ in range(count):
                key = self.prop_value(key_type, key_struct_type, key_path)
                value = self.prop_value(value_type, value_struct_type, value_path)
                values.append(
                    {
                        "key": key,
                        "value": value,
                    }
                )
            value = {
                "key_type": key_type,
                "value_type": value_type,
                "key_struct_type": key_struct_type,
                "value_struct_type": value_struct_type,
                "id": _id,
                "value": values,
            }
        else:
            raise Exception(f"Unknown type: {type_name} ({path})")
        value["type"] = type_name
        return value

    def prop_value(self, type_name: str, struct_type_name: str, path: str):
        if type_name == "StructProperty":
            return self.struct_value(struct_type_name, path)
        elif type_name == "EnumProperty":
            return self.fstring()
        elif type_name == "NameProperty":
            return self.fstring()
        elif type_name == "IntProperty":
            return self.i32()
        elif type_name == "BoolProperty":
            return self.bool()
        else:
            raise Exception(f"Unknown property value type: {type_name} ({path})")

    def struct(self, path: str) -> dict[str, Any]:
        struct_type = self.fstring()
        struct_id = self.guid()
        _id = self.optional_guid()
        value = self.struct_value(struct_type, path)
        return {
            "struct_type": struct_type,
            "struct_id": struct_id,
            "id": _id,
            "value": value,
        }

    def struct_value(self, struct_type: str, path: str = ""):
        if struct_type == "Vector":
            return {
                "x": self.double(),
                "y": self.double(),
                "z": self.double(),
            }
        elif struct_type == "DateTime":
            return self.u64()
        elif struct_type == "Guid":
            return self.guid()
        elif struct_type == "Quat":
            return {
                "x": self.double(),
                "y": self.double(),
                "z": self.double(),
                "w": self.double(),
            }
        elif struct_type == "LinearColor":
            return {
                "r": self.float(),
                "g": self.float(),
                "b": self.float(),
                "a": self.float(),
            }
        else:
            if self.debug:
                print(f"Assuming struct type: {struct_type} ({path})")
            return self.properties_until_end(path)

    def array_property(self, array_type: str, size: int, path: str):
        count = self.u32()
        value = {}
        if array_type == "StructProperty":
            prop_name = self.fstring()
            prop_type = self.fstring()
            self.u64()
            type_name = self.fstring()
            _id = self.guid()
            self.skip(1)
            prop_values = []
            for _ in range(count):
                prop_values.append(self.struct_value(type_name, f"{path}.{prop_name}"))
            value = {
                "prop_name": prop_name,
                "prop_type": prop_type,
                "values": prop_values,
                "type_name": type_name,
                "id": _id,
            }
        else:
            value = {
                "values": self.array_value(array_type, count, size, path),
            }
        return value

    def array_value(self, array_type: str, count: int, size: int, path: str):
        values = []
        decode_func: Callable
        if array_type == "EnumProperty":
            decode_func = self.fstring
        elif array_type == "NameProperty":
            decode_func = self.fstring
        elif array_type == "Guid":
            decode_func = self.guid
        elif array_type == "ByteProperty":
            if size == count:
                # Special case this and read faster in one go
                return self.byte_list(count)
            else:
                raise Exception("Labelled ByteProperty not implemented")
        else:
            raise Exception(f"Unknown array type: {array_type} ({path})")
        for _ in range(count):
            values.append(decode_func())

        return values

    def compressed_short_rotator(self) -> tuple[float, float, float]:
        short_pitch = self.u16() if self.bool() else 0
        short_yaw = self.u16() if self.bool() else 0
        short_roll = self.u16() if self.bool() else 0
        pitch = short_pitch * (360.0 / 65536.0)
        yaw = short_yaw * (360.0 / 65536.0)
        roll = short_roll * (360.0 / 65536.0)
        return [pitch, yaw, roll]

    def serializeint(self, component_bit_count: int) -> int:
        b = bytearray(self.read((component_bit_count + 7) // 8))
        if (component_bit_count % 8) != 0:
            b[-1] &= (1 << (component_bit_count % 8)) - 1
        value = int.from_bytes(b, "little")
        return value

    def packed_vector(self, scale_factor: int) -> tuple[float, float, float]:
        component_bit_count_and_extra_info = self.u32()
        component_bit_count = component_bit_count_and_extra_info & 63
        extra_info = component_bit_count_and_extra_info >> 6
        if component_bit_count > 0:
            x = self.serializeint(component_bit_count)
            y = self.serializeint(component_bit_count)
            z = self.serializeint(component_bit_count)
            sign_bit = 1 << (component_bit_count - 1)
            x = (x & (sign_bit - 1)) - (x & sign_bit)
            y = (y & (sign_bit - 1)) - (y & sign_bit)
            z = (z & (sign_bit - 1)) - (z & sign_bit)

            if extra_info:
                x /= scale_factor
                y /= scale_factor
                z /= scale_factor
            return (x, y, z)
        else:
            received_scaler_type_size = 8 if extra_info else 4
            if received_scaler_type_size == 8:
                x = self.double()
                y = self.double()
                z = self.double()
                return (x, y, z)
            else:
                x = self.float()
                y = self.float()
                z = self.float()
                return (x, y, z)

    def ftransform(self) -> dict[str, dict[str, float]]:
        return {
            "rotation": {
                "x": self.double(),
                "y": self.double(),
                "z": self.double(),
                "w": self.double(),
            },
            "translation": {
                "x": self.double(),
                "y": self.double(),
                "z": self.double(),
            },
            "scale3d": {
                "x": self.double(),
                "y": self.double(),
                "z": self.double(),
            },
        }


def uuid_writer(writer, s: Union[str, uuid.UUID, UUID]):
    if isinstance(s, str):
        s = uuid.UUID(s)
    if isinstance(s, uuid.UUID):
        b = s.bytes
        ub = bytes(
            [
                b[0x3],
                b[0x2],
                b[0x1],
                b[0x0],
                b[0x7],
                b[0x6],
                b[0x5],
                b[0x4],
                b[0xB],
                b[0xA],
                b[0x9],
                b[0x8],
                b[0xF],
                b[0xE],
                b[0xD],
                b[0xC],
            ]
        )
    elif isinstance(s, UUID):
        ub = s.raw_bytes
    writer.write(ub)


def instance_id_writer(writer, d):
    uuid_writer(writer, d["guid"])
    uuid_writer(writer, d["instance_id"])


class FArchiveWriter:
    data: io.BytesIO
    size: int
    custom_properties: dict[str, tuple[Callable, Callable]]
    debug: bool

    def __init__(
        self,
        custom_properties: dict[str, tuple[Callable, Callable]] = {},
        debug: bool = os.environ.get("DEBUG", "0") == "1",
    ):
        self.data = io.BytesIO()
        self.custom_properties = custom_properties
        self.debug = debug

    def __enter__(self):
        self.data.seek(0)
        return self

    def __exit__(self, type, value, traceback):
        self.data.close()

    def copy(self) -> "FArchiveWriter":
        return FArchiveWriter(self.custom_properties)

    def bytes(self) -> bytes:
        pos = self.data.tell()
        self.data.seek(0)
        b = self.data.read()
        self.data.seek(pos)
        return b

    def write(self, data: bytes):
        self.data.write(data)

    def bool(self, bool: bool):
        self.data.write(struct.pack("?", bool))

    def fstring(self, string: str) -> int:
        start = self.data.tell()
        if string == "":
            self.i32(0)
        elif string.isascii():
            str_bytes = string.encode("ascii")
            self.i32(len(str_bytes) + 1)
            self.data.write(str_bytes)
            self.data.write(b"\x00")
        else:
            str_bytes = string.encode("utf-16-le", errors="surrogatepass")
            assert len(str_bytes) % 2 == 0
            self.i32(-((len(str_bytes) // 2) + 1))
            self.data.write(str_bytes)
            self.data.write(b"\x00\x00")
        return self.data.tell() - start

    def i16(self, i: int):
        self.data.write(struct.pack("h", i))

    def u16(self, i: int):
        self.data.write(struct.pack("H", i))

    def i32(self, i: int):
        self.data.write(struct.pack("i", i))

    def u32(self, i: int):
        self.data.write(struct.pack("I", i))

    def i64(self, i: int):
        self.data.write(struct.pack("q", i))

    def u64(self, i: int):
        self.data.write(struct.pack("Q", i))

    def float(self, i: float):
        self.data.write(struct.pack("f", i))

    def double(self, i: float):
        self.data.write(struct.pack("d", i))

    def byte(self, b: int):
        self.data.write(bytes([b]))

    def u(self, b: int):
        self.data.write(struct.pack("B", b))

    def guid(self, u: Union[str, uuid.UUID, UUID]):
        uuid_writer(self, u)

    def optional_guid(self, u: Optional[Union[str, uuid.UUID, UUID]]):
        if u is None:
            self.bool(False)
        else:
            self.bool(True)
            uuid_writer(self, u)

    def tarray(
        self, type_writer: Callable[["FArchiveWriter", dict[str, Any]], None], array
    ):
        self.u32(len(array))
        for i in range(len(array)):
            type_writer(self, array[i])

    def properties(self, properties: dict[str, Any]):
        for key in properties:
            self.fstring(key)
            self.property(properties[key])
        self.fstring("None")

    def property(self, property: dict[str, Any]):
        # write type_name
        self.fstring(property["type"])
        nested_writer = self.copy()
        size: int
        property_type = property["type"]
        size = nested_writer.property_inner(property_type, property)
        buf = nested_writer.bytes()
        # write size
        self.u64(size)
        self.write(buf)

    def property_inner(self, property_type: str, property: dict[str, Any]) -> int:
        if "custom_type" in property:
            if property["custom_type"] in self.custom_properties:
                size = self.custom_properties[property["custom_type"]][1](
                    self, property_type, property
                )
            else:
                raise Exception(
                    f"Unknown custom property type: {property['custom_type']}"
                )
        elif property_type == "StructProperty":
            size = self.struct(property)
        elif property_type == "IntProperty":
            self.optional_guid(property.get("id", None))
            self.i32(property["value"])
            size = 4
        elif property_type == "Int64Property":
            self.optional_guid(property.get("id", None))
            self.i64(property["value"])
            size = 8
        elif property_type == "FixedPoint64Property":
            self.optional_guid(property.get("id", None))
            self.i32(property["value"])
            size = 4
        elif property_type == "FloatProperty":
            self.optional_guid(property.get("id", None))
            self.float(property["value"])
            size = 4
        elif property_type == "StrProperty":
            self.optional_guid(property.get("id", None))
            size = self.fstring(property["value"])
        elif property_type == "NameProperty":
            self.optional_guid(property.get("id", None))
            size = self.fstring(property["value"])
        elif property_type == "EnumProperty":
            self.fstring(property["value"]["type"])
            self.optional_guid(property.get("id", None))
            size = self.fstring(property["value"]["value"])
        elif property_type == "BoolProperty":
            self.bool(property["value"])
            self.optional_guid(property.get("id", None))
            size = 0
        elif property_type == "ArrayProperty":
            self.fstring(property["array_type"])
            self.optional_guid(property.get("id", None))
            array_writer = self.copy()
            array_writer.array_property(property["array_type"], property["value"])
            array_buf = array_writer.bytes()
            size = len(array_buf)
            self.write(array_buf)
        elif property_type == "MapProperty":
            self.fstring(property["key_type"])
            self.fstring(property["value_type"])
            self.optional_guid(property.get("id", None))
            map_writer = self.copy()
            map_writer.u32(0)
            map_writer.u32(len(property["value"]))
            for entry in property["value"]:
                map_writer.prop_value(
                    property["key_type"], property["key_struct_type"], entry["key"]
                )
                map_writer.prop_value(
                    property["value_type"],
                    property["value_struct_type"],
                    entry["value"],
                )
            map_buf = map_writer.bytes()
            size = len(map_buf)
            self.write(map_buf)
        else:
            raise Exception(f"Unknown property type: {property_type}")
        return size

    def struct(self, property: dict[str, Any]) -> int:
        self.fstring(property["struct_type"])
        self.guid(property["struct_id"])
        self.optional_guid(property.get("id", None))
        start = self.data.tell()
        self.struct_value(property["struct_type"], property["value"])
        return self.data.tell() - start

    def struct_value(self, struct_type: str, value):
        if struct_type == "Vector":
            self.double(value["x"])
            self.double(value["y"])
            self.double(value["z"])
        elif struct_type == "DateTime":
            self.u64(value)
        elif struct_type == "Guid":
            self.guid(value)
        elif struct_type == "Quat":
            self.double(value["x"])
            self.double(value["y"])
            self.double(value["z"])
            self.double(value["w"])
        elif struct_type == "LinearColor":
            self.float(value["r"])
            self.float(value["g"])
            self.float(value["b"])
            self.float(value["a"])
        else:
            if self.debug:
                print(f"Assuming struct type: {struct_type}")
            return self.properties(value)

    def prop_value(self, type_name: str, struct_type_name: str, value):
        if type_name == "StructProperty":
            self.struct_value(struct_type_name, value)
        elif type_name == "EnumProperty":
            self.fstring(value)
        elif type_name == "NameProperty":
            self.fstring(value)
        elif type_name == "IntProperty":
            self.i32(value)
        elif type_name == "BoolProperty":
            self.bool(value)
        else:
            raise Exception(f"Unknown property value type: {type_name}")

    def array_property(self, array_type: str, value: dict[str, Any]):
        count = len(value["values"])
        self.u32(count)
        if array_type == "StructProperty":
            self.fstring(value["prop_name"])
            self.fstring(value["prop_type"])
            nested_writer = self.copy()
            for i in range(count):
                nested_writer.struct_value(value["type_name"], value["values"][i])
            data_buf = nested_writer.bytes()
            self.u64(len(data_buf))
            self.fstring(value["type_name"])
            self.guid(value["id"])
            self.u(0)
            self.write(data_buf)
        else:
            self.array_value(array_type, count, value["values"])

    def array_value(self, array_type: str, count: int, values: list[Any]):
        for i in range(count):
            if array_type == "IntProperty":
                self.i32(values[i])
            elif array_type == "Int64Property":
                self.i64(values[i])
            elif array_type == "FloatProperty":
                self.float(values[i])
            elif array_type == "StrProperty":
                self.fstring(values[i])
            elif array_type == "NameProperty":
                self.fstring(values[i])
            elif array_type == "EnumProperty":
                self.fstring(values[i])
            elif array_type == "BoolProperty":
                self.bool(values[i])
            elif array_type == "ByteProperty":
                self.byte(values[i])
            else:
                raise Exception(f"Unknown array type: {array_type}")

    def compressed_short_rotator(self, pitch: float, yaw: float, roll: float):
        short_pitch = round(pitch * (65536.0 / 360.0)) & 0xFFFF
        short_yaw = round(yaw * (65536.0 / 360.0)) & 0xFFFF
        short_roll = round(roll * (65536.0 / 360.0)) & 0xFFFF
        if short_pitch != 0:
            self.bool(True)
            self.u16(short_pitch)
        else:
            self.bool(False)
        if short_yaw != 0:
            self.bool(True)
            self.u16(short_yaw)
        else:
            self.bool(False)
        if short_roll != 0:
            self.bool(True)
            self.u16(short_roll)
        else:
            self.bool(False)

    @staticmethod
    def unreal_round_float_to_int(value: float) -> int:
        return int(value)

    @staticmethod
    def unreal_get_bits_needed(value: int) -> int:
        massaged_value = value ^ (value >> 63)
        return 65 - FArchiveWriter.count_leading_zeroes(massaged_value)

    @staticmethod
    def count_leading_zeroes(value: int) -> int:
        return 67 - len(bin(-value)) & ~value >> 64

    def serializeint(self, component_bit_count: int, value: int):
        self.write(
            int.to_bytes(value, (component_bit_count + 7) // 8, "little", signed=True)
        )

    def packed_vector(self, scale_factor: int, x: float, y: float, z: float):
        max_exponent_for_scaling = 52
        max_value_to_scale = 1 << max_exponent_for_scaling
        max_exponent_after_scaling = 62
        max_scaled_value = 1 << max_exponent_after_scaling
        scaled_x = x * scale_factor
        scaled_y = y * scale_factor
        scaled_z = z * scale_factor
        if max(abs(scaled_x), abs(scaled_y), abs(scaled_z)) < max_scaled_value:
            use_scaled_value = min(abs(x), abs(y), abs(z)) < max_value_to_scale
            if use_scaled_value:
                x = self.unreal_round_float_to_int(scaled_x)
                y = self.unreal_round_float_to_int(scaled_y)
                z = self.unreal_round_float_to_int(scaled_z)
            else:
                x = self.unreal_round_float_to_int(x)
                y = self.unreal_round_float_to_int(y)
                z = self.unreal_round_float_to_int(z)

            component_bit_count = max(
                self.unreal_get_bits_needed(x),
                self.unreal_get_bits_needed(y),
                self.unreal_get_bits_needed(z),
            )
            component_bit_count_and_scale_info = (
                1 << 6 if use_scaled_value else 0
            ) | component_bit_count
            self.u32(component_bit_count_and_scale_info)
            self.serializeint(component_bit_count, x)
            self.serializeint(component_bit_count, y)
            self.serializeint(component_bit_count, z)
        else:
            component_bit_count = 0
            component_bit_count_and_scale_info = (1 << 6) | component_bit_count
            self.u32(component_bit_count_and_scale_info)
            self.double(x)
            self.double(y)
            self.double(z)

    def ftransform(self, value: dict[str, dict[str, float]]):
        self.double(value["rotation"]["x"])
        self.double(value["rotation"]["y"])
        self.double(value["rotation"]["z"])
        self.double(value["rotation"]["w"])
        self.double(value["translation"]["x"])
        self.double(value["translation"]["y"])
        self.double(value["translation"]["z"])
        self.double(value["scale3d"]["x"])
        self.double(value["scale3d"]["y"])
        self.double(value["scale3d"]["z"])



import zlib

MAGIC_BYTES = b"PlZ"


def decompress_sav_to_gvas(data: bytes) -> tuple[bytes, int]:
    uncompressed_len = int.from_bytes(data[0:4], byteorder="little")
    compressed_len = int.from_bytes(data[4:8], byteorder="little")
    magic_bytes = data[8:11]
    save_type = data[11]
    # Check for magic bytes
    if magic_bytes != MAGIC_BYTES:
        if (
            magic_bytes == b"\x00\x00\x00"
            and uncompressed_len == 0
            and compressed_len == 0
        ):
            raise Exception(
                f"not a compressed Palworld save, found too many null bytes, this is likely corrupted"
            )
        raise Exception(
            f"not a compressed Palworld save, found {magic_bytes} instead of {MAGIC_BYTES}"
        )
    # Valid save types
    if save_type not in [0x30, 0x31, 0x32]:
        raise Exception(f"unknown save type: {save_type}")
    # We only have 0x31 (single zlib) and 0x32 (double zlib) saves
    if save_type not in [0x31, 0x32]:
        raise Exception(f"unhandled compression type: {save_type}")
    if save_type == 0x31:
        # Check if the compressed length is correct
        if compressed_len != len(data) - 12:
            raise Exception(f"incorrect compressed length: {compressed_len}")
    # Decompress file
    uncompressed_data = zlib.decompress(data[12:])
    if save_type == 0x32:
        # Check if the compressed length is correct
        if compressed_len != len(uncompressed_data):
            raise Exception(f"incorrect compressed length: {compressed_len}")
        # Decompress file
        uncompressed_data = zlib.decompress(uncompressed_data)
    # Check if the uncompressed length is correct
    if uncompressed_len != len(uncompressed_data):
        raise Exception(f"incorrect uncompressed length: {uncompressed_len}")

    return uncompressed_data, save_type


def compress_gvas_to_sav(data: bytes, save_type: int) -> bytes:
    uncompressed_len = len(data)
    compressed_data = zlib.compress(data)
    compressed_len = len(compressed_data)
    if save_type == 0x32:
        compressed_data = zlib.compress(compressed_data)

    # Create a byte array and append the necessary information
    result = bytearray()
    result.extend(uncompressed_len.to_bytes(4, byteorder="little"))
    result.extend(compressed_len.to_bytes(4, byteorder="little"))
    result.extend(MAGIC_BYTES)
    result.extend(bytes([save_type]))
    result.extend(compressed_data)

    return bytes(result)



"""An interactive console."""

from getpass import getpass
from typing import Type

from rcon.client import BaseClient
from rcon.config import Config
from rcon.exceptions import SessionTimeout, WrongPassword


__all__ = ['PROMPT', 'rconcmd']


EXIT_COMMANDS = {'exit', 'quit'}
MSG_LOGIN_ABORTED = '\nLogin aborted. Bye.'
MSG_EXIT = '\nBye.'
MSG_SESSION_TIMEOUT = 'Session timed out. Please login again.'
PROMPT = 'RCON {host}:{port}> '


def read_host() -> str:
    """Reads the host."""

    while True:
        try:
            return input('Host: ')
        except KeyboardInterrupt:
            print()
            continue


def read_port() -> int:
    """Reads the port."""

    while True:
        try:
            port = input('Port: ')
        except KeyboardInterrupt:
            print()
            continue

        try:
            port = int(port)
        except ValueError:
            print(f'Invalid integer: {port}')
            continue

        if 0 <= port <= 65535:
            return port

        print(f'Invalid port: {port}')


def read_passwd() -> str:
    """Reads the password."""

    while True:
        try:
            return getpass('Password: ')
        except KeyboardInterrupt:
            print()


def get_config(host: str, port: int, passwd: str) -> Config:
    """Reads the necessary arguments."""

    if host is None:
        host = read_host()

    if port is None:
        port = read_port()

    if passwd is None:
        passwd = read_passwd()

    return Config(host, port, passwd)


def login(client: BaseClient, passwd: str) -> str:
    """Performs a login."""

    while True:
        try:
            client.login(passwd)
        except WrongPassword:
            print('Wrong password.')
            passwd = read_passwd()
            continue

        return passwd


def process_input(client: BaseClient, passwd: str, prompt: str) -> bool:
    """Processes the CLI input."""

    try:
        command = input(prompt)
    except KeyboardInterrupt:
        print()
        return True
    except EOFError:
        print(MSG_EXIT)
        return False

    try:
        command, *args = command.split()
    except ValueError:
        return True

    if command in EXIT_COMMANDS:
        return False

    try:
        result = client.run(command, *args)
    except SessionTimeout:
        print(MSG_SESSION_TIMEOUT)

        try:
            login(client, passwd)
        except EOFError:
            print(MSG_LOGIN_ABORTED)
            return False

        return True

    if result:
        print(result)

    return True


def rconcmd(
        client_cls: Type[BaseClient], host: str, port: int, passwd: str, *,
        prompt: str = PROMPT
):
    """Initializes the console."""

    try:
        host, port, passwd = get_config(host, port, passwd)
    except EOFError:
        print(MSG_EXIT)
        return

    prompt = prompt.format(host=host, port=port)

    with client_cls(host, port) as client:
        try:
            passwd = login(client, passwd)
        except EOFError:
            print(MSG_LOGIN_ABORTED)
            return

        while True:
            if not process_input(client, passwd, prompt):
                break



"""Wrapper for readline support."""

from logging import Logger
from pathlib import Path

try:
    from readline import read_history_file, write_history_file
except ModuleNotFoundError:
    read_history_file = write_history_file = lambda _: None


__all__ = ['CommandHistory']


HIST_FILE = Path.home().joinpath('.rconshell_history')


class CommandHistory:
    """Context manager for the command line history."""

    __slots__ = ('logger',)

    def __init__(self, logger: Logger):
        """Sets the logger to use."""
        self.logger = logger

    def __enter__(self):
        """Loads the history file."""
        try:
            read_history_file(HIST_FILE)
        except FileNotFoundError:
            self.logger.warning(
                'Could not find history file: %s', HIST_FILE
            )
        except PermissionError:
            self.logger.error(
                'Insufficient permissions to read: %s', HIST_FILE
            )

        return self

    def __exit__(self, *_):
        """Writes to the history file."""
        try:
            write_history_file(HIST_FILE)
        except PermissionError:
            self.logger.error(
                'Insufficient permissions to write: %s', HIST_FILE
            )



EmptySkillObject = {
    "array_type": "NameProperty",
    "id": None,
    "value": {
        "values": []
    },
    "type": "ArrayProperty"
}

EmptyLevelObject = {
    "id": None,
    "value": 1,
    "type": "IntProperty"
}

EmptyRankObject = {
    "id": None,
    "value": 1,
    "type": "IntProperty"
}

EmptyMeleeObject = {
    "id": None,
    "value": 0,
    "type": "IntProperty"
}

EmptyShotObject = {
    "id": None,
    "value": 0,
    "type": "IntProperty"
}

EmptyDefenceObject = {
    "id": None,
    "value": 0,
    "type": "IntProperty"
}

EmptyExpObject = {
    "id": None,
    "value": 0,
    "type": "IntProperty"
}

EmptyRarePalObject = {
    "value": False,
    "id": None,
    "type": "BoolProperty"
}
    
EmptyMovesObject = {
    "array_type": "EnumProperty",
    "id": None,
    "value": {
        "values": []
    },
    "type": "ArrayProperty"
}



from typing import Any, Sequence

from lib.archive import *


def decode(
    reader: FArchiveReader, type_name: str, size: int, path: str
) -> dict[str, Any]:
    if type_name != "ArrayProperty":
        raise Exception(f"Expected ArrayProperty, got {type_name}")
    value = reader.property(type_name, size, path, nested_caller_path=path)
    data_bytes = value["value"]["values"]
    value["value"] = decode_bytes(data_bytes)
    return value


def decode_bytes(m_bytes: Sequence[int]) -> dict[str, Any]:
    reader = FArchiveReader(bytes(m_bytes), debug=False)
    data = {}
    data["instance_id"] = reader.guid()
    data["concrete_model_instance_id"] = reader.guid()
    data["base_camp_id_belong_to"] = reader.guid()
    data["group_id_belong_to"] = reader.guid()
    data["hp"] = {
        "current": reader.i32(),
        "max": reader.i32(),
    }
    data["initital_transform_cache"] = reader.ftransform()
    data["repair_work_id"] = reader.guid()
    data["owner_spawner_level_object_instance_id"] = reader.guid()
    data["owner_instance_id"] = reader.guid()
    data["build_player_uid"] = reader.guid()
    data["interact_restrict_type"] = reader.byte()
    data["stage_instance_id_belong_to"] = {
        "id": reader.guid(),
        "valid": reader.u32() > 0,
    }
    data["created_at"] = reader.i64()
    if not reader.eof():
        raise Exception("Warning: EOF not reached")
    return data


def encode(
    writer: FArchiveWriter, property_type: str, properties: dict[str, Any]
) -> int:
    if property_type != "ArrayProperty":
        raise Exception(f"Expected ArrayProperty, got {property_type}")
    del properties["custom_type"]
    encoded_bytes = encode_bytes(properties["value"])
    properties["value"] = {"values": [b for b in encoded_bytes]}
    return writer.property_inner(property_type, properties)


def encode_bytes(p: dict[str, Any]) -> bytes:
    writer = FArchiveWriter()

    writer.guid(p["instance_id"])
    writer.guid(p["concrete_model_instance_id"])
    writer.guid(p["base_camp_id_belong_to"])
    writer.guid(p["group_id_belong_to"])

    writer.i32(p["hp"]["current"])
    writer.i32(p["hp"]["max"])

    writer.ftransform(p["initital_transform_cache"])

    writer.guid(p["repair_work_id"])
    writer.guid(p["owner_spawner_level_object_instance_id"])
    writer.guid(p["owner_instance_id"])
    writer.guid(p["build_player_uid"])

    writer.byte(p["interact_restrict_type"])

    writer.guid(p["stage_instance_id_belong_to"]["id"])
    writer.u32(1 if p["stage_instance_id_belong_to"]["valid"] else 0)

    writer.i64(p["created_at"])

    encoded_bytes = writer.bytes()
    return encoded_bytes



"""GTK based GUI."""

from argparse import ArgumentParser, Namespace
from json import dump, load
from logging import DEBUG, INFO, basicConfig, getLogger
from os import getenv, name
from pathlib import Path
from socket import gaierror, timeout
from typing import Iterable, NamedTuple, Type

from gi import require_version
require_version('Gtk', '3.0')
from gi.repository import Gtk

from rcon import battleye, source
from rcon.client import BaseClient
from rcon.config import LOG_FORMAT
from rcon.exceptions import SessionTimeout, WrongPassword


__all__ = ['main']


if name == 'posix':
    CACHE_DIR = Path.home().joinpath('.cache')
elif name == 'nt':
    CACHE_DIR = Path(getenv('TEMP') or getenv('TMP'))
else:
    raise NotImplementedError('Unsupported operating system.')


CACHE_FILE = CACHE_DIR.joinpath('rcongui.json')
LOGGER = getLogger('rcongui')


def get_args() -> Namespace:
    """Parses the command line arguments."""

    parser = ArgumentParser(description='A minimalistic, GTK-based RCON GUI.')
    parser.add_argument(
        '-B', '--battleye', action='store_true',
        help='use BattlEye RCon instead of Source RCON'
    )
    parser.add_argument(
        '-d', '--debug', action='store_true',
        help='print additional debug information'
    )
    parser.add_argument(
        '-t', '--timeout', type=float, metavar='seconds',
        help='connection timeout in seconds'
    )
    return parser.parse_args()


class RCONParams(NamedTuple):
    """Represents the RCON parameters."""

    host: str
    port: int
    passwd: str
    command: Iterable[str]


class GUI(Gtk.Window):  # pylint: disable=R0902
    """A GTK based GUI for RCON."""

    def __init__(self, args: Namespace):
        """Initializes the GUI."""
        super().__init__(title='RCON GUI')
        self.args = args

        self.set_position(Gtk.WindowPosition.CENTER)

        self.grid = Gtk.Grid()
        self.add(self.grid)

        self.host = Gtk.Entry()
        self.host.set_placeholder_text('Host')
        self.grid.attach(self.host, 0, 0, 1, 1)

        self.port = Gtk.SpinButton.new_with_range(0, 65535, 1)
        self.port.set_placeholder_text('Port')
        self.grid.attach(self.port, 1, 0, 1, 1)

        self.passwd = Gtk.Entry()
        self.passwd.set_placeholder_text('Password')
        self.passwd.set_visibility(False)
        self.grid.attach(self.passwd, 2, 0, 1, 1)

        self.command = Gtk.Entry()
        self.command.set_placeholder_text('Command')
        self.grid.attach(self.command, 0, 1, 2, 1)

        self.button = Gtk.Button(label='Run')
        self.button.connect('clicked', self.on_button_clicked)
        self.grid.attach(self.button, 2, 1, 1, 1)

        self.result = Gtk.TextView()
        self.result.set_wrap_mode(Gtk.WrapMode.WORD)
        self.result.set_property('editable', False)
        self.grid.attach(self.result, 0, 2, 2, 1)

        self.savepw = Gtk.CheckButton(label='Save password')
        self.grid.attach(self.savepw, 2, 2, 1, 1)

        self.load_gui_settings()

    @property
    def client_cls(self) -> Type[BaseClient]:
        """Returns the client class."""
        return battleye.Client if self.args.battleye else source.Client

    @property
    def result_text(self) -> str:
        """Returns the result text."""
        if (buf := self.result.get_buffer()) is not None:
            start = buf.get_iter_at_line(0)
            end = buf.get_iter_at_line(buf.get_line_count())
            return buf.get_text(start, end, True)

        return ''

    @result_text.setter
    def result_text(self, text: str):
        """Sets the result text."""
        if (buf := self.result.get_buffer()) is not None:
            buf.set_text(text)

    @property
    def gui_settings(self) -> dict:
        """Returns the GUI settings as a dict."""
        json = {
            'host': self.host.get_text(),
            'port': self.port.get_value_as_int(),
            'command': self.command.get_text(),
            'result': self.result_text,
            'savepw': (savepw := self.savepw.get_active())
        }

        if savepw:
            json['passwd'] = self.passwd.get_text()

        return json

    @gui_settings.setter
    def gui_settings(self, json: dict):
        """Sets the GUI settings."""
        self.host.set_text(json.get('host', ''))
        self.port.set_value(json.get('port', 0))
        self.passwd.set_text(json.get('passwd', ''))
        self.command.set_text(json.get('command', ''))
        self.result_text = json.get('result', '')
        self.savepw.set_active(json.get('savepw', False))

    def load_gui_settings(self) -> None:
        """Loads the GUI settings from the cache file."""
        try:
            with CACHE_FILE.open('rb') as cache:
                self.gui_settings = load(cache)
        except FileNotFoundError:
            LOGGER.warning('Cache file not found: %s', CACHE_FILE)
        except PermissionError:
            LOGGER.error('Insufficient permissions to read: %s', CACHE_FILE)
        except ValueError:
            LOGGER.error('Cache file contains garbage: %s', CACHE_FILE)

    def save_gui_settings(self):
        """Saves the GUI settings to the cache file."""
        try:
            with CACHE_FILE.open('wb') as cache:
                dump(self.gui_settings, cache)
        except PermissionError:
            LOGGER.error('Insufficient permissions to read: %s', CACHE_FILE)

    def show_error(self, message: str):
        """Shows an error message."""
        message_dialog = Gtk.MessageDialog(
            transient_for=self,
            message_type=Gtk.MessageType.ERROR,
            buttons=Gtk.ButtonsType.OK,
            text=message
        )
        message_dialog.run()
        message_dialog.destroy()

    def run_rcon(self) -> str:
        """Returns the current RCON settings."""
        with self.client_cls(
                self.host.get_text().strip(),
                self.port.get_value_as_int(),
                timeout=self.args.timeout,
                passwd=self.passwd.get_text()
        ) as client:
            return client.run(*self.command.get_text().strip().split())

    def on_button_clicked(self, _):
        """Runs the client."""
        try:
            result = self.run_rcon()
        except ValueError as error:
            self.show_error(str(error))
        except gaierror as error:
            self.show_error(error.strerror)
        except ConnectionRefusedError:
            self.show_error('Connection refused.')
        except (TimeoutError, timeout):
            self.show_error('Connection timed out.')
        except WrongPassword:
            self.show_error('Wrong password.')
        except SessionTimeout:
            self.show_error('Session timed out.')
        else:
            self.result_text = result

    def terminate(self, *args, **kwargs):
        """Saves the settings and terminates the application."""
        self.save_gui_settings()
        Gtk.main_quit(*args, **kwargs)


def main() -> None:
    """Starts the GUI."""

    args = get_args()
    basicConfig(format=LOG_FORMAT, level=DEBUG if args.debug else INFO)
    win = GUI(args)
    win.connect('destroy', win.terminate)
    win.show_all()
    Gtk.main()



from typing import Any, Sequence

from lib.archive import *

# def decode_map_concrete_model(
#     reader: FArchiveReader, type_name: str, size: int, path: str
# ) -> dict[str, Any]:
#     if type_name != "StructProperty":
#         raise Exception(f"Expected StructProperty, got {type_name}")
#     value = reader.property(type_name, size, path, nested_caller_path=path)
#     # Decode the raw bytes for the map object and replace the raw data
#     raw_bytes = value["value"]["RawData"]["value"]["values"]
#     print("".join(f"{b:02x}" for b in raw_bytes))
#     # value["value"]["RawData"]["value"] = decode_map_concrete_model_bytes(raw_bytes)
#     # Decode the raw bytes for the module map and replace the raw data
#     # group_map = value["value"]
#     # for group in group_map:
#     #     group_type = group["value"]["GroupType"]["value"]["value"]
#     #     group_bytes = group["value"]["RawData"]["value"]["values"]
#     #     group["value"]["RawData"]["value"] = decode_map_concrete_model_bytes(
#     #         group_bytes, group_type
#     #     )
#     # EPalMapObjectConcreteModelModuleType::None = 0,
#     # EPalMapObjectConcreteModelModuleType::ItemContainer = 1,
#     # EPalMapObjectConcreteModelModuleType::CharacterContainer = 2,
#     # EPalMapObjectConcreteModelModuleType::Workee = 3,
#     # EPalMapObjectConcreteModelModuleType::Energy = 4,
#     # EPalMapObjectConcreteModelModuleType::StatusObserver = 5,
#     # EPalMapObjectConcreteModelModuleType::ItemStack = 6,
#     # EPalMapObjectConcreteModelModuleType::Switch = 7,
#     # EPalMapObjectConcreteModelModuleType::PlayerRecord = 8,
#     # EPalMapObjectConcreteModelModuleType::BaseCampPassiveEffect = 9,
#     # EPalMapObjectConcreteModelModuleType::PasswordLock = 10,
#     return value


# def decode_map_concrete_model_bytes(m_bytes: Sequence[int]) -> dict[str, Any]:
#     if len(m_bytes) == 0:
#         return None
#     reader = FArchiveReader(bytes(m_bytes), debug=False)
#     map_concrete_model = {}

#     if not reader.eof():
#         raise Exception("Warning: EOF not reached")
#     return map_concrete_model


# def encode_map_concrete_model(
#     writer: FArchiveWriter, property_type: str, properties: dict[str, Any]
# ) -> int:
#     if property_type != "MapProperty":
#         raise Exception(f"Expected MapProperty, got {property_type}")
#     del properties["custom_type"]
#     # encoded_bytes = encode_map_concrete_model_bytes(properties["value"]["RawData"]["value"])
#     # properties["value"]["RawData"]["value"] = {"values": [b for b in encoded_bytes]}
#     return writer.property_inner(property_type, properties)


# def encode_map_concrete_model_bytes(p: dict[str, Any]) -> bytes:
#     writer = FArchiveWriter()

#     encoded_bytes = writer.bytes()
#     return encoded_bytes



"""Common errors handler."""

from logging import Logger
from socket import timeout

from rcon.exceptions import ConfigReadError
from rcon.exceptions import SessionTimeout
from rcon.exceptions import UserAbort
from rcon.exceptions import WrongPassword


__all__ = ['ErrorHandler']


ERRORS = {
    UserAbort: (1, None),
    ConfigReadError: (2, None),
    ConnectionRefusedError: (3, 'Connection refused.'),
    (TimeoutError, timeout): (4, 'Connection timed out.'),
    WrongPassword: (5, 'Wrong password.'),
    SessionTimeout: (6, 'Session timed out.')
}


class ErrorHandler:
    """Handles common errors and exits."""

    __slots__ = ('logger', 'exit_code')

    def __init__(self, logger: Logger):
        """Sets the logger."""
        self.logger = logger
        self.exit_code = 0

    def __enter__(self):
        return self

    def __exit__(self, _, value: Exception, __):
        """Checks for connection errors and exits respectively."""
        if value is None:
            return True

        for typ, (exit_code, message) in ERRORS.items():
            if isinstance(value, typ):
                self.exit_code = exit_code

                if message:
                    self.logger.error(message)

                return True

        return None
